@@ -53,7 +53,7 @@
   (progn
     (setq *break-on-signals* nil)
     (setq sb-kernel::*current-error-depth* 0))
-  (setq *stack-top-hint* nil))
+  (setq sb-debug:*stack-top-hint* nil))
 
 (defun !printer-control-init ()
   (setq *print-readably* nil
@@ -63,6 +63,7 @@
         *print-radix* nil
         *print-vector-length* nil
         *print-circle* nil
+        *print-circle-not-shared* nil
         *print-case* :upcase
         *print-array* t
         *print-gensym* t
@@ -116,14 +117,13 @@
                                       :if-does-not-exist nil)
                 (when stream
                   (let ((*package* (find-package "SB-KERNEL"))) (read stream))))))
+    (dolist (item list)
+      (destructuring-bind (object hash) item
+        (let ((calc (sxhash object)))
+          (unless (= hash calc)
+            (error "cross-compiler SXHASH failure on ~S: ~X vs ~X" object hash calc)))))
     (when list
-      (dolist (item list)
-        (destructuring-bind (object hash) item
-          (unless (= (sxhash object) hash)
-            (error "SXHASH computed wrong answer for ~S. Got ~x should be ~x"
-                   object hash (sxhash object)))))
-      (format t "~&cross-compiler SXHASH tests passed: ~D cases~%"
-              (length list)))))
+      (format t "~&cross-compiler SXHASH tests passed: ~D cases~%" (length list)))))
 
 ;;; called when a cold system starts up
 (defun !cold-init ()
@@ -156,9 +156,7 @@
   ;; this to be initialized, so we initialize it right away.
   (show-and-call !random-cold-init)
 
-  ;; We can't do such much as a simple PROCLAIM without this global
-  ;; hash-table (because of WARN-IF-INLINE-FAILED/PROCLAIM)
-  (setf sb-c::*emitted-full-calls* (make-hash-table :test 'equal :synchronized t))
+  (setq sb-c::*compilation-unit* nil) ; its DEFVAR is not processed yet
 
   ;; All sorts of things need INFO and/or (SETF INFO).
   (/show0 "about to SHOW-AND-CALL !GLOBALDB-COLD-INIT")
@@ -169,10 +167,14 @@
   ;; And now *CURRENT-THREAD*
   (sb-thread::init-main-thread)
 
+  (show-and-call !hash-table-cold-init)
+
   ;; not sure why this is needed on some architectures. Dark magic.
+  #-linkage-space
   (setf (fdefn-fun (find-or-create-fdefn '%coerce-callable-for-call))
         #'%coerce-callable-to-fun)
   (show-and-call !loader-cold-init)
+  #+linkage-space (show-and-call sb-vm::!initialize-lisp-linkage-table)
   ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
   ;; It was fine if T because in that case the legality of the arg is certain.
   ;; And be extra paranoid - ensure that it really gets called.
@@ -199,6 +201,13 @@
   ;; Must be done before any non-opencoded array references are made.
   (show-and-call sb-vm::!hairy-data-vector-reffer-init)
 
+  ;; It's not unreasonable to call PARSE-INTEGER in cold-init, which demands that
+  ;; WHITESPACE[1]P work properly, so it needs *STANDARD-READTABLE* established.
+  (let ((*readtable* (make-readtable)))
+    (show-and-call !reader-cold-init)
+    ;; *STANDARD-READTABLE* is assigned last to avoid an error about altering it.
+    (setf *standard-readtable* *readtable*))
+
   ;; Various toplevel forms call MAKE-ARRAY, which calls SUBTYPEP, so
   ;; the basic type machinery needs to be initialized before toplevel
   ;; forms run.
@@ -255,12 +264,6 @@
 
   (makunbound '*!cold-toplevels*) ; so it gets GC'd
 
-  ;; Need the static-space replica of the assembly routine jump vector
-  ;; filled in, and the static space vector of static fdefns.
-  ;; This matters only for code that gets compiled to dynamic space,
-  ;; so it's OK that it occurs somewhat late in cold-init.
-  #+x86-64 (sb-vm::validate-asm-routine-vector)
-
   #+win32 (show-and-call reinit-internal-real-time)
 
   ;; Set sane values again, so that the user sees sane values instead
@@ -277,8 +280,15 @@
 
   ;; We run through queued-up type and ftype proclaims that were made
   ;; before the type system was initialized, and (since it is now
-  ;; initalized) reproclaim them..
-  (mapcar #'proclaim sb-c::*queued-proclaims*)
+  ;; initalized) reproclaim them.
+  (loop for claim in sb-c::*queued-proclaims*
+        do
+        (when (eq (car claim) 'ftype)
+          ;; Avoid warnings about mismatched types.
+          (loop for name in (cddr claim)
+                do (setf (info :function :where-from name) :assumed)))
+        (proclaim claim))
+
   (makunbound 'sb-c::*queued-proclaims*)
 
   (show-and-call os-cold-init-or-reinit)
@@ -301,9 +311,6 @@
     (show-and-call !reader-cold-init)
     (show-and-call !sharpm-cold-init)
     (show-and-call !backq-cold-init)
-    ;; The *STANDARD-READTABLE* is assigned at last because the above
-    ;; functions would operate on the standard readtable otherwise---
-    ;; which would result in an error.
     (setf *standard-readtable* *readtable*))
   (setf *readtable* (copy-readtable *standard-readtable*))
   (setf sb-debug:*debug-readtable* (copy-readtable *standard-readtable*))
@@ -403,7 +410,6 @@ process to continue normally."
     ;; can be called, as pretty much anything can assume that it is set.
     (when total ; newly started process, and not a failed save attempt
       (sb-thread::init-main-thread)
-      #+x86-64 (sb-vm::validate-asm-routine-vector)
       (rebuild-package-vector))
     ;; Initialize streams next, so that any errors can be printed
     (stream-reinit t)
