@@ -350,9 +350,13 @@
            (bignum-truncate-single-digit (make-small-bignum number) divisor)
            (bignum-truncate (make-small-bignum number) divisor)))
       ((ratio (or float rational))
-       (let ((q (truncate (numerator number)
-                          (* (denominator number) divisor))))
-         (values q (- number (* q divisor)))))
+       (if (eql divisor 1)
+           (let ((q (truncate (numerator number)
+                              (denominator number))))
+             (values q (- number q)))
+           (let ((q (truncate (numerator number)
+                              (* (denominator number) divisor))))
+             (values q (- number (* q divisor))))))
       ((bignum fixnum)
        (bignum-truncate-single-digit number divisor))
       ((bignum bignum)
@@ -416,10 +420,30 @@
         (if (single-digit-bignum-p divisor)
             (bignum-truncate-single-digit (make-small-bignum number) divisor)
             (bignum-truncate (make-small-bignum number) divisor))))
-      ((ratio (or float rational))
-       (let ((q (truncate (numerator number)
-                          (* (denominator number) divisor))))
-         (fixup (values q (- number (* q divisor))))))
+      ((ratio integer)
+       (let ((numerator (numerator number)))
+         (if (eql divisor 1)
+             (let ((q (truncate numerator (denominator number))))
+               (if (< numerator 0)
+                   (let ((q-1 (- q 1)))
+                     (values q-1 (- number q-1)))
+                   (values q (- number q))))
+             (let ((q (truncate numerator (* (denominator number) divisor))))
+               (if (if (minusp divisor)
+                       (> numerator 0)
+                       (< numerator 0))
+                   (let ((q-1 (- q 1)))
+                     (values q-1 (- number (* q-1 divisor))))
+                   (values q (- number (* q divisor))))))))
+      ((ratio (or float ratio))
+       (let* ((q (truncate (numerator number) (* (denominator number) divisor)))
+              (mult (* q divisor)))
+         (if (if (minusp divisor)
+                 (< mult number)
+                 (> mult number))
+             (let ((q-1 (- q 1)))
+               (values q-1 (- number (* q-1 divisor))))
+             (values q (- number (* q divisor))))))
       ((bignum fixnum)
        (fixup (bignum-truncate-single-digit number divisor)))
       ((bignum bignum)
@@ -480,10 +504,31 @@
         (if (single-digit-bignum-p divisor)
             (bignum-truncate-single-digit (make-small-bignum number) divisor)
             (bignum-truncate (make-small-bignum number) divisor))))
-      ((ratio (or float rational))
-       (let ((q (truncate (numerator number)
-                          (* (denominator number) divisor))))
-         (fixup (values q (- number (* q divisor))))))
+      ((ratio integer)
+       (let ((numerator (numerator number)))
+         (if (eql divisor 1)
+             (let ((q (truncate numerator (denominator number))))
+               (if (> numerator 0)
+                   (let ((q+1 (+ q 1)))
+                     (values q+1 (- number q+1)))
+                   (values q (- number q))))
+             (let ((q (truncate numerator (* (denominator number) divisor))))
+               (if (if (minusp divisor)
+                       (< numerator 0)
+                       (> numerator 0))
+                   (let ((q+1 (+ q 1)))
+                     (values q+1 (- number (* q+1 divisor))))
+                   (values q (- number (* q divisor))))))))
+      ((ratio (or float ratio))
+       (let* ((q (truncate (numerator number)
+                           (* (denominator number) divisor)))
+              (mult (* q divisor)))
+         (if (if (minusp divisor)
+                 (> mult number)
+                 (< mult number))
+             (let ((q+1 (+ q 1)))
+               (values q+1 (- number (* q+1 divisor))))
+             (values q (- number (* q divisor))))))
       ((bignum fixnum)
        (fixup (bignum-truncate-single-digit number divisor)))
       ((bignum bignum)
@@ -506,7 +551,7 @@
        (truncate-float double-float))
       (((foreach fixnum bignum ratio)
         (foreach single-float double-float #+long-float long-float))
-        (truncate-float (dispatch-type divisor))))))
+       (truncate-float (dispatch-type divisor))))))
 
 (defun truncate1 (number divisor)
   (declare (explicit-check)
@@ -568,12 +613,6 @@
     (multiple-value-bind (res rem) (,op number divisor)
       (values (float res (if (floatp rem) rem 1.0)) rem))))
 
-;;; Declare these guys inline to let them get optimized a little.
-;;; ROUND and FROUND are not declared inline since they seem too
-;;; obscure and too big to inline-expand by default. Also, this gives
-;;; the compiler a chance to pick off the unary float case.
-(declaim (inline fceiling ffloor ftruncate))
-
 #-round-float
 (defun fround (number &optional (divisor 1))
   "Same as ROUND, but returns first value as a float."
@@ -1742,15 +1781,15 @@ and the number of 0 bits if INTEGER is negative."
 
 ;;;; miscellaneous number predicates
 
-(macrolet ((def (name doc)
-             `(defun ,name (number) ,doc
+(macrolet ((def (name var doc)
+             `(defun ,name (,var) ,doc
                 (declare (explicit-check))
-                (,name number))))
-  (def zerop "Is this number zero?")
-  (def plusp "Is this real number strictly positive?")
-  (def minusp "Is this real number strictly negative?")
-  (def oddp "Is this integer odd?")
-  (def evenp "Is this integer even?"))
+                (,name ,var))))
+  (def zerop number "Is this number zero?")
+  (def plusp number "Is this real number strictly positive?")
+  (def minusp number "Is this real number strictly negative?")
+  (def oddp integer "Is this integer odd?")
+  (def evenp integer "Is this integer even?"))
 
 ;;;; modular functions
 #.
