@@ -53,6 +53,8 @@
 (declaim (type (and (vector t) (not simple-array)) *constraint-universe*))
 (defvar *constraint-universe*)
 (defvar *blocks-to-terminate*)
+(defvar *constraint-blocks*)
+(defvar *constraint-blocks-p*)
 
 (defstruct (vector-length-constraint
             (:constructor make-vector-length-constraint (var))
@@ -329,10 +331,9 @@
   (etypecase y
     (ctype
        (awhen (lambda-var-ctype-constraints x)
-         (dolist (con (gethash (sb-kernel::type-class y) it) nil)
+         (dolist (con (gethash y it) nil)
            (when (and (eq (constraint-kind con) kind)
-                      (eq (constraint-not-p con) not-p)
-                      (type= (constraint-y con) y))
+                      (eq (constraint-not-p con) not-p))
              (return-from find-constraint con)))
          nil))
     (lvar
@@ -373,7 +374,7 @@
       (ctype
        (let ((index (ensure-hash (lambda-var-ctype-constraints x)))
              (vec   (ensure-vec  (lambda-var-inheritable-constraints x))))
-         (push con (gethash (sb-kernel::type-class y) index))
+         (push con (gethash y index))
          (vector-push-extend con vec)))
       (lvar
        (let ((index (ensure-hash (lambda-var-eq-constraints x))))
@@ -414,7 +415,7 @@
 (declaim (inline type-for-constraints-p))
 (defun type-for-constraints-p (type)
   (not (or (eq type *universal-type*)
-           (contains-hairy-type-p type))))
+           (opaque-type-p type))))
 
 ;;; Actual conset interface
 ;;;
@@ -667,7 +668,7 @@
 (defun add-combination-test-constraints (use constraints
                                          consequent-constraints
                                          alternative-constraints
-                                         quick-p)
+                                         quick-p &optional (nth-value 0))
   (flet ((add (fun lvar y &optional no-complement)
            (let ((x (ok-lvar-lambda-var lvar constraints)))
              (if no-complement
@@ -686,23 +687,28 @@
                                            alternative-constraints)))
          (prop (triples target)
            (map nil (lambda (constraint)
-                      (destructuring-bind (kind x y &optional not-p)
-                          constraint
-                        (when (and kind x y)
-                          (let ((x (if (lvar-p x)
-                                       (ok-lvar-lambda-var x constraints)
-                                       x)))
-                            (when x
-                              (add-test-constraint quick-p
-                                                   kind x y
-                                                   not-p constraints
-                                                   target))))))
+                      (cond ((not constraint))
+                            ((eq (car constraint) 'equality)
+                             (destructuring-bind (op x y) (cdr constraint)
+                               (add-equality-constraint op x y constraints target nil)))
+                            (t
+                             (destructuring-bind (kind x y &optional not-p)
+                                 constraint
+                               (when (and kind x y)
+                                 (let ((x (if (lvar-p x)
+                                              (ok-lvar-lambda-var x constraints)
+                                              x)))
+                                   (when x
+                                     (add-test-constraint quick-p
+                                                          kind x y
+                                                          not-p constraints
+                                                          target))))))))
                 triples)))
     (when (eq (combination-kind use) :known)
       (binding* ((info (combination-fun-info use) :exit-if-null)
                  (propagate (fun-info-constraint-propagate-if info) :exit-if-null))
         (multiple-value-bind (lvar type if else no-complement)
-            (funcall propagate use constraints)
+            (funcall propagate use constraints nth-value)
           (prop if consequent-constraints)
           (prop else alternative-constraints)
           (when (and lvar type)
@@ -741,9 +747,16 @@
                     (add 'typep (ok-lvar-lambda-var (ref-lvar node) constraints)
                          (specifier-type 'null) t)
                     (let ((use (principal-lvar-ref-use (ref-lvar use))))
-                      (when (and use
-                                 (not (ref-p use)))
-                        (process-node use))))
+                      (if use
+                          (unless (ref-p use)
+                            (process-node use))
+                          (multiple-value-bind (node nth-value) (mv-principal-lvar-ref-use (ref-lvar node))
+                            (when (combination-p node)
+                              (add-combination-test-constraints node constraints
+                                                                consequent-constraints
+                                                                alternative-constraints
+                                                                quick-p
+                                                                nth-value))))))
                    (combination
                     (unless (eq (combination-kind node) :error)
                       (let ((name (uncross
@@ -753,17 +766,20 @@
                         (add-equality-constraints name args
                                                   constraints consequent-constraints alternative-constraints)
                         (case name
-                          ((%typep %instance-typep)
+                          ((typep %typep %instance-typep)
                            (let ((type (second args)))
                              (when (constant-lvar-p type)
-                               (let ((val (lvar-value type)))
-                                 (add 'typep
-                                      (ok-lvar-lambda-var (first args) constraints)
-                                      (if (ctype-p val)
-                                          val
-                                          (let ((*compiler-error-context* node))
-                                            (specifier-type val)))
-                                      nil)))))
+                               (let* ((val (lvar-value type))
+                                      (ctype (if (ctype-p val)
+                                                 val
+                                                 (careful-specifier-type val))))
+                                 (when (and ctype
+                                            (type-for-constraints-p ctype))
+                                   (add 'typep
+                                        (ok-lvar-lambda-var (first args) constraints)
+                                        ctype
+                                        nil
+                                        (first args)))))))
                           ((eq eql)
                            (let* ((arg1 (first args))
                                   (var1 (ok-lvar-lambda-var arg1 constraints))
@@ -906,14 +922,16 @@
                     (sb-xc:> (type-bound-number ref) (type-bound-number x))))))
     (let* ((x-bound (bound x))
            (y-bound (exclude (bound y)))
-           (new-bound (cond ((not x-bound)
-                             y-bound)
-                            ((not y-bound)
-                             x-bound)
-                            ((tighter-p y-bound x-bound)
-                             y-bound)
-                            (t
-                             x-bound))))
+           (new-bound (coerce-for-bound
+                       (cond ((not x-bound)
+                              y-bound)
+                             ((not y-bound)
+                              x-bound)
+                             ((tighter-p y-bound x-bound)
+                              y-bound)
+                             (t
+                              x-bound))
+                       (numeric-type-format x))))
       (if greater
           (modified-numeric-type x :low new-bound)
           (modified-numeric-type x :high new-bound)))))
@@ -1021,7 +1039,7 @@
                      (setf not-xset (alloc-xset)))
                    (add-to-xset x not-xset))))
            (intersect-result (other-type)
-             (setf type (type-approx-intersection2 type other-type))))
+             (setf type (type-intersection type other-type))))
       (declare (inline intersect-result))
       (do-propagatable-constraints (con (constraints variable))
         (let* ((kind (constraint-kind con))
@@ -1096,24 +1114,38 @@
 (defun type-after-comparison (operator not-p current-type type)
   (case operator
     ((= eq)
-     (unless not-p
-       (multiple-value-bind (lo hi)
-           (if (numeric-type-p type)
-               (values (numeric-type-low type)
-                       (numeric-type-high type))
-               ;; Doesn't handle infinities
-               (let ((int (type-approximate-interval type)))
-                 (and int
-                      (values
-                       (interval-low int)
-                       (interval-high int)))))
-         (when (or lo hi)
-           (type-intersection current-type
-                              (type-union (make-numeric-type :low lo
-                                                             :high hi)
-                                          (make-numeric-type :complexp :complex
-                                                             :low lo
-                                                             :high hi)))))))
+     (multiple-value-bind (lo hi)
+         (if (numeric-type-p type)
+             (values (numeric-type-low type)
+                     (numeric-type-high type))
+             ;; Doesn't handle infinities
+             (let ((int (type-approximate-interval type)))
+               (and int
+                    (values
+                     (interval-low int)
+                     (interval-high int)))))
+       (if not-p
+           (when (and (csubtypep current-type (specifier-type 'integer))
+                      (integerp hi)
+                      (eql lo hi))
+             ;; Cut off an adjacent bound if it's not EQ
+             (multiple-value-bind (c-lo)
+                 (let ((int (type-approximate-interval current-type)))
+                   (and int
+                        (values
+                         (interval-low int)
+                         (interval-high int))))
+               (when (and c-lo
+                          (= hi c-lo))
+                 (type-intersection current-type
+                                    (make-numeric-type :low (1+ lo))))))
+           (when (or lo hi)
+             (type-intersection current-type
+                                (type-union (make-numeric-type :low lo
+                                                               :high hi)
+                                            (make-numeric-type :complexp :complex
+                                                               :low lo
+                                                               :high hi)))))))
     (t
      (multiple-value-bind (greater equal)
          (if not-p
@@ -1153,19 +1185,10 @@
       (when type
         (unless set
           (setf (lambda-var-unused-initial-value leaf) nil))
-        ;; CHANGE-CLASS can change the type, lower down to standard-object,
+        ;; CHANGE-CLASS can change the type,
         ;; type propagation for classes is not as important anyway.
-        (cond #-sb-xc-host
-              ((and
-                (eq sb-pcl::**boot-state** 'sb-pcl::complete)
-                (block nil
-                  (let ((standard-object (find-classoid 'standard-object)))
-                    (sb-kernel::map-type
-                     (lambda (type)
-                       (when (and (classoid-p type)
-                                  (csubtypep type standard-object))
-                         (return t)))
-                     type)))))
+        (cond ((logtest sb-kernel::ctype-contains-class
+                        (sb-kernel::type-flags type)))
               (t
                (derive-node-type ref
                                  (make-single-value-type type))
@@ -1179,7 +1202,13 @@
         ;; Find unchanged eql refs to a set variable.
         (when (lambda-var-sets leaf)
           (let (mark
-                (eq (lambda-var-eq-constraints leaf)))
+                (old-mark (ref-same-refs ref))
+                (eq (lambda-var-eq-constraints leaf))
+                ;; It's cheap to compute but not cheap to reoptimize
+                ;; everything if it's not needed.
+                ;; Some transforms ask for same-leaf-ref-p after :ir1-phases.
+                (compute (lambda-var-compute-same-refs leaf))
+                reoptimize)
             (when eq
               (loop for other-ref in (leaf-refs leaf)
                     unless (eq other-ref ref)
@@ -1189,9 +1218,14 @@
                            (unless mark
                              (setf mark (list 0))
                              (setf (ref-same-refs ref) mark))
+                           (when (and compute
+                                       (not (and old-mark
+                                                 (eq old-mark (ref-same-refs other-ref)))))
+                             (setf reoptimize t)
+                             (reoptimize-lvar (node-lvar other-ref)))
                            (setf (ref-same-refs other-ref) mark)))))
-            (when mark
-              (reoptimize-node ref))))))))
+            (when reoptimize
+              (reoptimize-lvar (node-lvar ref)))))))))
 
 ;;;; Flow analysis
 
@@ -1225,15 +1259,22 @@
        (let ((fun (bind-lambda node)))
          (functional-kind-case fun
            (let
-            (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
-                  for var in (lambda-vars fun)
-                  and val in (combination-args call)
-                  when (and val (lambda-var-constraints var))
-                  do (let ((type (lvar-type val)))
-                       (when (type-for-constraints-p type)
-                         (conset-add-constraint gen 'typep var type nil)))
-                     (maybe-add-eql-var-var-constraint var val gen)
-                     (add-var-result-constraints var val gen)))
+               (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
+                     for var in (lambda-vars fun)
+                     and val in (combination-args call)
+                     when (and val (lambda-var-constraints var))
+                     do (let ((type (lvar-type val)))
+                          (when (type-for-constraints-p type)
+                            (conset-add-constraint gen 'typep var type nil)))
+                        (maybe-add-eql-var-var-constraint var val gen)
+                        (add-var-result-constraints var val gen)))
+           ((nil)
+            (loop for var in (lambda-vars fun)
+                  for type = (leaf-defined-type var)
+                  do
+                  (when (and (lambda-var-constraints var)
+                             (type-for-constraints-p type))
+                    (conset-add-constraint gen 'typep var type nil))))
            (mv-let
             (add-mv-let-result-constraints (lvar-dest (node-lvar (first (lambda-refs fun)))) fun gen)))))
       (ref
@@ -1241,6 +1282,7 @@
          (maybe-add-eql-var-lvar-constraint node gen)
          (when preprocess-refs-p
            (constrain-ref-type node gen))))
+      (delay)
       (cast
        (let* ((lvar (cast-value node))
               (var (ok-lvar-lambda-var lvar gen))
@@ -1307,7 +1349,7 @@
                                  (conset= call-in gen))))
               (setf (combination-constraints-in node)
                     (copy-conset gen))
-              (when (boundp '*constraint-blocks*)
+              (when *constraint-blocks-p*
                 (enqueue-block-for-constraints (lambda-block fun))))))))))
   gen)
 
@@ -1395,10 +1437,7 @@
                  (when (block-type-check (lambda-block fun))
                    ;; This is optimistic, make sure it's going to be
                    ;; processed.
-                   (setf (lambda-var-unused-initial-value var) t))
-                 (loop for ref in (lambda-var-refs var)
-
-                       do (setf (ref-same-refs ref) nil))))))
+                   (setf (lambda-var-unused-initial-value var) t))))))
       (frob fun)
       (dolist (let (lambda-lets fun))
         (frob let)))))
@@ -1561,8 +1600,6 @@
     (when (eql (car x) obj)
       (return-from nconc-new list))))
 
-(defvar *constraint-blocks*)
-
 (defun enqueue-block-for-constraints (block)
   (when (block-type-check block)
     (setq *constraint-blocks* (nconc-new block *constraint-blocks*))))
@@ -1587,7 +1624,8 @@
     ;; done, hence any inherited type constraints from such
     ;; constraints will be wrong as well.
     (dolist (join-types-p '(nil t))
-      (let ((*constraint-blocks* (copy-list rest-of-blocks)))
+      (let ((*constraint-blocks-p* t)
+            (*constraint-blocks* (copy-list rest-of-blocks)))
         ;; The rest of the blocks.
         (dolist (block rest-of-blocks)
           (aver (eq block (pop *constraint-blocks*)))
@@ -1618,10 +1656,15 @@
         (setf (if-alternative-constraints last) nil)
         (setf (if-consequent-constraints last) nil))))
 
-  (let (*blocks-to-terminate*)
+  (let (*blocks-to-terminate*
+        *constraint-blocks-p*)
     (dolist (block (find-and-propagate-constraints component))
       (unless (block-delete-p block)
         (use-result-constraints block)))
+    #+sb-devel
+    (when (and *compiler-trace-output*
+               (memq :constraints *compile-trace-targets*))
+      (print-constraints component))
     (loop for node in *blocks-to-terminate*
           do (maybe-terminate-block node nil)))
   (values))
